/* vi: set sw=4 ts=4: */
/*
 * Utility routine.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <syslog.h>
#include <openssl/crypto.h>
#include <openssl/err.h>

#define USE_NETAPP_DES 1
#define ENABLE_OPENSSL 1
#define FIPS_MODE_ENABLE		1
#define FIPS_MODE_DSIABLE		0

#if ENABLE_USE_BB_CRYPT

static int i64c(int i)
{
	i &= 0x3f;
	if (i == 0)
		return '.';
	if (i == 1)
		return '/';
	if (i < 12)
		return ('0' - 2 + i);
	if (i < 38)
		return ('A' - 12 + i);
	return ('a' - 38 + i);
}


static char*
to64(char *s, unsigned v, int n)
{
	while (--n >= 0) {
		/* *s++ = ascii64[v & 0x3f]; */
		*s++ = i64c(v);
		v >>= 6;
	}
	return s;
}


/*
 * DES and MD5 crypt implementations are taken from uclibc.
 * They were modified to not use static buffers.
 */
/* Common for them */
static const uint8_t ascii64[] = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

#include "pw_encrypt_des.c"
#include "pw_encrypt_md5.c"
void fips_enable(void);
#if ENABLE_OPENSSL
#include "pw_encrypt_sha_openssl.c"
#else
#if ENABLE_USE_BB_CRYPT_SHA
#include "pw_encrypt_sha.c"
#endif
#endif

void fips_enable(void) {
	int fips_mode = -1;
	int ret = -1;
	unsigned long err = 0;

	fips_mode = FIPS_mode();
	if (fips_mode == FIPS_MODE_DSIABLE) {
		ret = FIPS_mode_set(FIPS_MODE_ENABLE);
		if(ret == FIPS_MODE_ENABLE)
		{
			syslog(LOG_INFO, "Succeeded to enable FIPS mode.");
		} else {
			err = ERR_get_error();
			syslog(LOG_INFO, "Failed to enable FIPS mode(%lx, %s).",err, ERR_error_string(err, NULL));
		}
	}
}

/* Other advanced crypt ids: */
/* $2$ or $2a$: Blowfish */
/* $5$: SHA-256 */
/* $6$: SHA-512 */
/* TODO: implement SHA - http://people.redhat.com/drepper/SHA-crypt.txt */

static struct const_des_ctx *des_cctx;
static struct des_ctx *des_ctx;

/* my_crypt returns malloc'ed data */
static char *my_crypt(const char *key, const char *salt)
{
	/* MD5 or SHA? */
	if (salt[0] == '$' && salt[1] && salt[2] == '$') {
		if (salt[1] == '1')
			return md5_crypt(xzalloc(MD5_OUT_BUFSIZE), (unsigned char*)key, (unsigned char*)salt);

#if ENABLE_OPENSSL
		if (salt[1] == '5' || salt[1] == '6') {
			fips_enable();
			return sha_crypt_openssl((char*)key, (char*)salt);
		}
#else
#if ENABLE_USE_BB_CRYPT_SHA
		if (salt[1] == '5' || salt[1] == '6')
			return sha_crypt((char*)key, (char*)salt);
#endif
#endif
	}
#if USE_NETAPP_DES
	/* Must use NetApp specific Extende DES for SP root users and naroot users */
    return xstrdup(crypt(key, salt));
#endif
}


/* So far nobody wants to have it public */
static void my_crypt_cleanup(void)
{
	free(des_cctx);
	free(des_ctx);
	des_cctx = NULL;
	des_ctx = NULL;
}

char* FAST_FUNC pw_encrypt(const char *clear, const char *salt, int cleanup)
{
	char *encrypted;

	encrypted = my_crypt(clear, salt);

	if (cleanup)
		my_crypt_cleanup();

	return encrypted;
}

#else /* if !ENABLE_USE_BB_CRYPT */

char* FAST_FUNC pw_encrypt(const char *clear, const char *salt, int cleanup)
{
	return xstrdup(crypt(clear, salt));
}

#endif
