
/*
 * Copyright (c) 2009 - Network Appliance Inc.
 *
 */

/*********************************************************************
 * VSC452 driver module for SPI manual mode interface
 *
 *********************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>
#include <asm/ioctl.h>

#include "spi_vsc.h"

/*****Prototye Functions*************************/
static  int  vsc452_spi_init (void );
static void  vsc452_spi_exit (void);
static  int  vsc452_spi_open (struct inode *, struct file *);
static  int  vsc452_spi_release (struct inode *, struct file *);
static  int  vsc452_spi_read (struct file *, char *, size_t, loff_t *);
static  int  vsc452_spi_write (struct file *, const char *, size_t, loff_t *);
static  long  vsc452_spi_ioctl ( struct file *, uint, ulong);

static struct file_operations vsc452_spi_ops;
static void  *spi_dev;

static int vsc452_spi_manual_mode_write_read(vsc452_spi_command_list_t *commands, vsc452_spi_configuration_t *spi_cfg) ;
vsc452_spi_configuration_t g_spi_cfg ;

int __init
vsc452_spi_init(void)
{
    int  err, ret=0;
    uint32_t spi_base;
    void *dev= &spi_dev;

    printk(" VSC452 SPI Init \n ");

    memset(&vsc452_spi_ops,0,sizeof(vsc452_spi_ops));
    vsc452_spi_ops.unlocked_ioctl   = vsc452_spi_ioctl;
    vsc452_spi_ops.open    = vsc452_spi_open;
    vsc452_spi_ops.release = vsc452_spi_release;
    vsc452_spi_ops.read    = vsc452_spi_read;
    vsc452_spi_ops.write   = vsc452_spi_write;
    vsc452_spi_ops.owner   = THIS_MODULE;

    /*********Register Character Device ********************/
    if ((err = register_chrdev(SPI_MAJOR_NO, SPI_NAME, &vsc452_spi_ops)) < 0) {
        printk ("failed to register device (err %d)\n", err);
        ret = -ENODEV;
        goto fail;
    }  else {
        printk ("registered device %s)\n", SPI_NAME);
    }

    /* Remap registers */
    spi_base = SPI_BASE_OFFSET;
    if ((dev = ioremap_nocache(spi_base, SPI_SIZE)) == NULL) {
        printk("failed to map SPI IO space %08x-%08x to memory",spi_base, spi_base + SPI_SIZE - 1);
        ret = -EIO;
        goto fail;
    }

    /* check spi memory region */
    if ((err = check_mem_region(spi_base, SPI_SIZE)) < 0)  {
        printk("SPI IO space %08x-%08x already in use (err %d)",spi_base, spi_base + SPI_SIZE - 1, err);
        ret = -EBUSY;
        goto fail;
    }

    /* request for the memory region */
    if (request_mem_region(spi_base, SPI_SIZE, SPI_NAME)  == NULL) {
        printk("IO space %08x-%08x spi request mem region failed (err %d)",spi_base, spi_base + SPI_SIZE - 1, err);
        ret = -ENODEV;
        goto fail;
    }

    printk("SPI Module was sucessully loaded \n");
    return ret;

fail:
    vsc452_spi_exit();
    return ret;
}

static void
vsc452_spi_exit(void)
{
    void *dev= &spi_dev;
    uint32_t spi_base;
    printk(" VSC452 SPI Exit \n ");
    spi_base = SPI_BASE_OFFSET;

    /* release the memory occupied */
    release_mem_region(spi_base, SPI_SIZE);

    /* unregister the device */
    unregister_chrdev(SPI_MAJOR_NO, SPI_NAME);

    /* Character Device unmapping   */
    if (dev)
    {
        iounmap(dev);
        dev = NULL;
    }

    printk ( "SPI Driver is unloaded sucessfully.\n");
    return;
}

inline void spi_write_reg(unsigned long addr, unsigned long val)
{
    *((volatile uint32_t*)addr) = val;
}

inline unsigned long spi_read_reg(unsigned long addr )
{
    return ((*(volatile uint32_t*)addr));
}

static  int
vsc452_spi_open (struct inode *inode, struct file *file)
{
    file->f_op =&vsc452_spi_ops;
    return 0;
}

static  int
vsc452_spi_release (struct inode *inode,struct file *file)
{
    return 0;
}

static  int
vsc452_spi_read (struct file *file, char *buffer,size_t size, loff_t *offset)
{
    return 0;
}

static  int
vsc452_spi_write (struct file *file, const char *buffer,
size_t size, loff_t * offset)
{
    return 0;
}

static long
vsc452_spi_ioctl ( struct file *file, uint cmd , ulong arg)
{
    spi_ioctl_t pkt;

    switch ( cmd ) {
        case SPI_READ_REG:
            if (copy_from_user((void*)&pkt, (void*)arg, sizeof(spi_ioctl_t)))
                return -EFAULT;

            pkt.data =  VSC452_SPI_READ_REG(pkt.reg) ;

            if (copy_to_user((void*)arg, (void*)&pkt, sizeof(spi_ioctl_t)))
                return -EFAULT;
            break;
        case SPI_WRITE_REG:
            if (copy_from_user((void*)&pkt, (void*)arg, sizeof(spi_ioctl_t)))
                return -EFAULT;
            // WRITE_REG(pkt.reg, pkt.data);
            VSC452_SPI_WRITE_REG(pkt.reg,pkt.data) ;
            break;
        case SPI_SET_CFG:
            if (copy_from_user((void*)&g_spi_cfg, (void*)arg, sizeof(vsc452_spi_configuration_t)))
                return -EFAULT;
            break ;
        case SPI_MANUAL_WRITE:
        case SPI_MANUAL_READ:
        {
            vsc452_spi_command_list_t command_list ;
            vsc452_spi_command_cfg_t cmd_cfg[MAX_CMDS_IN_LIST] ;
            int ret_val = 0 ;
            if (copy_from_user(&command_list, (vsc452_spi_command_list_t *)arg, sizeof(vsc452_spi_configuration_t)))
                return -EFAULT;

            if ( command_list.num_cmds > MAX_CMDS_IN_LIST )
                return -EFAULT;

            if (copy_from_user(&cmd_cfg[0], command_list.cmd_details, ( command_list.num_cmds * sizeof(vsc452_spi_command_cfg_t))))
                return -EFAULT;

            command_list.cmd_details = &cmd_cfg[0] ;

            ret_val = vsc452_spi_manual_mode_write_read(&command_list, &g_spi_cfg) ;
            if ( ret_val != 0 ) {
                printk(" vsc452_spi_manual_mode_write_read returned error \n") ;
                return ret_val ;
            }

            if ( cmd == SPI_MANUAL_READ ) {
                if (copy_to_user((((vsc452_spi_command_list_t *)(arg))->cmd_details),
                                   (void*)&cmd_cfg[0],
                                   (command_list.num_cmds * sizeof(vsc452_spi_command_cfg_t))))
                return -EFAULT;
            }
            break ;
        }
        default:
            break;
	} /* End of switch */
    return 0;
}

#if 0
static void
vsc452_spi_dump_regs(char *str)
{
    printk(" ***** %s**** \n", str);
    printk(" ***** (00) Read reg    = 0x%x ***** \n",  (unsigned int)(VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_DATA_READ))) ;
    printk(" ***** (04) Write reg   = 0x%x ***** \n",  (unsigned int)(VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_DATA_WRITE))) ;
    printk(" ***** (08) Config reg  = 0x%x ***** \n",  (unsigned int)(VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_CONTROL))) ;
    printk(" ***** (0C) Status reg  = 0x%x ***** \n",  (unsigned int)(VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_STATUS))) ;
    printk(" ***** (10) Enable reg  = 0x%x ***** \n",  (unsigned int)(VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_ENABLE))) ;
    printk(" ******************* \n");
    printk(" ******************* \n");
}
#endif

static void
vsc452_spi_release_manual_mode_chip_select(void)
{
    /* Write 1 to release manual mode chip select bit. */
    VSC452_SPI_SET_FIELD(VSC452_SPI_MANUAL_MODE_STATUS,
                         VSC452_SPI_MANUAL_MODE_STATUS_RELEASE_MANUAL_MODE_CHIP_SELECT_MASK,
                         VSC452_SPI_MANUAL_MODE_STATUS_RELEASE_MANUAL_MODE_CHIP_SELECT_SHIFT,
                         1)  ;
}

static int
vsc452_spi_data_read(unsigned int *val)
{
    if ( val == NULL )
        return -EINVAL ;

    *val = VSC452_SPI_READ_REG(VSC452_SPI_MANUAL_MODE_DATA_READ) ;
    return 0 ;
}


static int
vsc452_spi_data_write(unsigned int val)
{
    if ( val > VSC452_SPI_MANUAL_DATA_WRITE_VAL_MAX )
        return -EINVAL ;

    VSC452_SPI_WRITE_REG(VSC452_SPI_MANUAL_MODE_DATA_WRITE,val) ;

    return 0 ;
}


static void
vsc452_spi_disable_manual_mode(void)
{
    VSC452_SPI_CLEAR_FIELD(VSC452_SPI_MANUAL_MODE_ENABLE,
                         VSC452_SPI_MANUAL_MODE_ENABLE_MASK,
                         VSC452_SPI_MANUAL_MODE_ENABLE_SHIFT) ;
}


static void
vsc452_spi_enable_manual_mode(void)
{
    VSC452_SPI_SET_FIELD(VSC452_SPI_MANUAL_MODE_ENABLE,
                         VSC452_SPI_MANUAL_MODE_ENABLE_MASK,
                         VSC452_SPI_MANUAL_MODE_ENABLE_SHIFT,
                         1) ;
}

static int
vsc452_spi_configure_chip_select(unsigned int val)
{
    if ( val > VSC452_SPI_MANUAL_CHIP_SELECT_VAL_MAX )
        return -EINVAL ;

    VSC452_SPI_SET_FIELD(VSC452_SPI_MANUAL_MODE_CONTROL,
                         VSC452_SPI_MANUAL_MODE_CONTROL_CHIP_SELECT_MASK,
                         VSC452_SPI_MANUAL_MODE_CONTROL_CHIP_SELECT_SHIFT,
                         val) ;
    return 0 ;
}

static int
vsc452_spi_configure_speed_select(unsigned int val)
{
    if ( val > VSC452_SPI_MANUAL_SPEED_SELECT_VAL_MAX )
        return -EINVAL ;

    VSC452_SPI_SET_FIELD(VSC452_SPI_MANUAL_MODE_CONTROL,
                         VSC452_SPI_MANUAL_MODE_CONTROL_SPEED_SELECT_MASK,
                         VSC452_SPI_MANUAL_MODE_CONTROL_SPEED_SELECT_SHIFT,
                         val) ;
    return 0 ;
}

static int
vsc452_spi_poll_bit(unsigned int addr, unsigned int mask, unsigned int shift, unsigned int poll_val, unsigned int cnt, unsigned int delay )
{
    unsigned int i = 0 ;
    volatile unsigned int val = 0 ;

    val = VSC452_SPI_READ_FIELD(addr, mask, shift) ;
    while ( ( val != poll_val ) && ( i < cnt ) )  {
        val = VSC452_SPI_READ_FIELD(addr, mask, shift) ;
        udelay(delay) ;
        i++ ;
    }

    if ( i < cnt ) {
        return 0 ;  /* Success */
    } else {
        return -EPERM ; /* Failure */
    }
}

static int
vsc452_spi_poll_for_manual_mode_active_bit(unsigned int val, unsigned int poll_cnt, unsigned int poll_delay)
{
    if ( val > VSC452_SPI_MANUAL_MODE_STATUS_MANUAL_MODE_ACTIVE_VAL_MAX )
        return -EINVAL ;

    return ( vsc452_spi_poll_bit(VSC452_SPI_MANUAL_MODE_STATUS,
                                 VSC452_SPI_MANUAL_MODE_STATUS_MANUAL_MODE_ACTIVE_MASK,
                                 VSC452_SPI_MANUAL_MODE_STATUS_MANUAL_MODE_ACTIVE_SHIFT,
                                 val, poll_cnt, poll_delay) );
}


static int
vsc452_spi_poll_for_next_byte_ready_bit(unsigned int val, unsigned int poll_cnt, unsigned int poll_delay)
{
    if ( val > VSC452_SPI_MANUAL_MODE_STATUS_NEXT_BYTE_READY_VAL_MAX )
        return -EINVAL ;

    return ( vsc452_spi_poll_bit(VSC452_SPI_MANUAL_MODE_STATUS,
                                 VSC452_SPI_MANUAL_MODE_STATUS_NEXT_BYTE_READY_MASK,
                                 VSC452_SPI_MANUAL_MODE_STATUS_NEXT_BYTE_READY_SHIFT,
                                 val, poll_cnt, poll_delay) );
}

static int
vsc452_spi_manual_mode_set_speed_chip_select_and_enable_manual_mode(vsc452_spi_configuration_t *spi_cfg)
{
    int ret_val = 0 ;

    if ( spi_cfg == NULL )
        return -EFAULT ;

    /* Configure speed select */
    if ( ( ret_val = vsc452_spi_configure_speed_select(spi_cfg->speed_select) ) == 0 ) {
        /* Configure chip select */
        if ( ( ret_val = vsc452_spi_configure_chip_select(spi_cfg->chip_select) ) == 0 ) {
            /* Enable Manual mode */
            vsc452_spi_enable_manual_mode() ;
        }
    }
    return ret_val ;
}

static int
vsc452_spi_config_width(unsigned int data_width_in_bits)
{
    unsigned int bit_cnt_val = 0 ;

    if ( ( data_width_in_bits > VSC452_SPI_MANUAL_BIT_COUNT_MAX ) || ( data_width_in_bits == 0 ) ) {
        return -EINVAL  ;
    }

    if ( data_width_in_bits == 8 )
        bit_cnt_val = 0 ;
    else
        bit_cnt_val = data_width_in_bits ;

    /* Now program the register. */
    VSC452_SPI_SET_FIELD(VSC452_SPI_MANUAL_MODE_CONTROL,
                         VSC452_SPI_MANUAL_MODE_CONTROL_BIT_COUNT_MASK,
                         VSC452_SPI_MANUAL_MODE_CONTROL_BIT_COUNT_SHIFT,
                         bit_cnt_val) ;
    return 0 ;
}

static int
vsc452_spi_manual_mode_write_read(vsc452_spi_command_list_t *commands, vsc452_spi_configuration_t *spi_cfg)
{
    unsigned int ret_val = 0 ;
    unsigned int i = 0 ;
    vsc452_spi_command_cfg_t *cmd_details ;

    if ( ( commands == NULL ) || ( spi_cfg == NULL ) )
        return -EFAULT ;

    /* Setup speed, chip select, and enable interface */
    if ( ( ret_val = vsc452_spi_manual_mode_set_speed_chip_select_and_enable_manual_mode(spi_cfg) ) != 0 )
        return ret_val ;

    /* Now send all the commands */
    for ( i = 0 ; i < commands->num_cmds ; i++ ) {
        cmd_details = commands->cmd_details++ ;
        if ( ( ret_val = vsc452_spi_config_width(cmd_details->width ) ) != 0 )
            return ret_val ;

        if ( ( ret_val = vsc452_spi_data_write(cmd_details->command) ) != 0 )
            return ret_val ;

        /* Poll for Nexy Byte ready bit to be 1 */
        if ( ( ret_val = vsc452_spi_poll_for_next_byte_ready_bit(1, spi_cfg->poll_cnt, spi_cfg->poll_delay) ) != 0 )
            return ret_val ;

        /* Wait before reading data. */
        udelay(10) ;

        if ( cmd_details->read_data_reg_after_command == VSC452_SPI_READ_AFTER_COMMAND ) {
            /* Now read data from spi read register */
            ret_val = vsc452_spi_data_read(&(cmd_details->read_val)) ;
            if ( ret_val != 0 )
                return ret_val ;
        }
    }

    /* Release manual mode chip select */
    vsc452_spi_release_manual_mode_chip_select() ;

    /* Poll for manual mode active bit to be 0 */
    ret_val = vsc452_spi_poll_for_manual_mode_active_bit(0 , spi_cfg->poll_cnt , spi_cfg->poll_delay) ;
    if ( ret_val != 0 )
        return ret_val ;

    /* Disable manual mode */
    vsc452_spi_disable_manual_mode() ;

    return ret_val ;
}

module_init(vsc452_spi_init);
module_exit(vsc452_spi_exit);

MODULE_DESCRIPTION("VSC452 SPI Driver");
MODULE_AUTHOR("NetApp Inc");
MODULE_LICENSE("GPL");
