/*
 * $Id$
 *
 * Copyright 2008, Network Appliance Inc.
 * Author: Jason McMullan <mcmullan@netapp.com>
 *
 */

#include <common.h>
#include <i2c.h>
#include <asm/io.h>

#include "vsc452/vsc452_i2c.h"

#define I2C_TIMEOUT	200	/* 200ms */

#define WAIT_UNTIL(condition, timeout_ms) \
	do { \
		int time_left = timeout_ms * 1000; \
		while (time_left > 0 && !(condition)) { \
			udelay((time_left > 100) ? 100 : time_left); \
			time_left -= (time_left > 100)  ? 100 : time_left; \
		}; \
	} while (0); if (!(condition))

int i2c_bus;
int shared_bus_seg=0;

static struct vsc452_i2c_bus {
	int speed;	/* offset into the speed map */
	uint32_t base;	/* Physical address */
} i2c[VSC452_I2C_INTERFACES];

static const struct {
	unsigned int speed;
	unsigned int select;
} i2c_speed_map[] = {
	{ .speed =   78000, .select = SID_CMS_CLK_78KHZ },
	{ .speed =   98000, .select = SID_CMS_CLK_98KHZ },
	{ .speed =  284000, .select = SID_CMS_CLK_284KHZ },
	{ .speed =  340000, .select = SID_CMS_CLK_340KHZ },
	{ .speed =  347000, .select = SID_CMS_CLK_347KHZ },
	{ .speed =  417000, .select = SID_CMS_CLK_417KHZ },
	{ .speed =  852000, .select = SID_CMS_CLK_852KHZ },
	{ .speed = 1042000, .select = SID_CMS_CLK_1042KHZ },
};

static inline uint32_t read32(uint32_t base)
{
	return *(volatile uint32_t *)(KSEG1ADDR(base));
}

static inline void write32(uint32_t base, uint32_t val)
{
	*(volatile uint32_t *)(KSEG1ADDR(base)) = val;
}

static inline uint32_t iread(int reg)
{
	return read32(i2c[i2c_bus].base + reg);
}

static inline void iwrite(int reg, uint32_t val)
{
	write32(i2c[i2c_bus].base + reg, val);
}

int i2c_set_bus_speed (unsigned int speed)
{
        int i;

        for (i = 0; i < ARRAY_SIZE(i2c_speed_map); i++) {
                if (speed <= i2c_speed_map[i].speed) {
                        break;
                }
        }

        /* Too fast? Then select the fastest possible.
	 */
        if (i == ARRAY_SIZE(i2c_speed_map)) {
                i--;
        }

        i2c[i2c_bus].speed = i;
        /* Set the speed, do not snoop the bus */
        iwrite(SIC, SIC_PIN | SIC_ES_CMS);
        iwrite(SID, i2c_speed_map[i2c[i2c_bus].speed].select);

        /* Return to data i/o mode */
        iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_ACK);

        return 0;
}

int i2c_set_bus_num (unsigned int bus)
{
	int i;

	if (bus >= VSC452_I2C_INTERFACES) {
		return -1;
	}

	i2c_bus = bus;
	i2c[i2c_bus].base = I2C_BASE(i2c_bus);

	/* if shared bus */
	if (i2c_bus == 2) {

		/* request shared bus */
		if (i2c_request_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}	
		
		/* wait for bus grant and if bus granted */
		if (i2c_grant_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}
		
		/* start short timer */
		fpga_start_short_timer();
	}
		
	/* Reset the controller */
	iwrite(SIBC, SIBC_SISR);
	iwrite(SIBC, SIBC_SCC | SIBC_SDC);

	/* Wiggle out any data on the line */
	if ((iread(SIBS) & SIBS_SDS) == 0) {
		for (i = 0; i < 8 * 16; i++) {
			iwrite(SIBC,        0 | SIBC_SDC);
			udelay(100);
			iwrite(SIBC, SIBC_SCC | SIBC_SDC);
			udelay(100);
		}
	}

	/* Master only */
	iwrite(SIC, SIC_PIN | SIC_ES_SCA);
	iwrite(SID, SID_SCA_ADDR(0x7F));

	/* Select the bus speed */
	i2c_set_bus_speed(98000); /* use only 98khz i2c bus speed */

	/* if shared bus */
	if (i2c_bus == 2) {
		/* release shared bus */
		i2c_release_shared_bus(shared_bus_seg);
	}
	
	return 0;
}

unsigned int i2c_get_bus_num (void)
{
	return i2c_bus;
}

int i2c_set_shared_bus_seg (unsigned int seg)
{
	if (seg > MAX_SEG_VAL) {
		return -1;
	}

	shared_bus_seg = seg;

	return 0;
}

unsigned int i2c_get_shared_bus_seg (void)
{
	return shared_bus_seg;
}


unsigned int i2c_get_bus_speed (void )
{
	return i2c_speed_map[i2c[i2c_bus].speed].speed;
}


#define I2C_STAT(tmp) \
	do { \
		debug("Status: 0x%08x (", tmp); \
		if ((tmp & SIC_PIN) == SIC_PIN) debug("PIN "); \
		if ((tmp & SIC_SSN) == SIC_SSN) debug("SSN "); \
		if ((tmp & SIC_STS) == SIC_STS) debug("STS "); \
		if ((tmp & SIC_BER) == SIC_BER) debug("BER "); \
		if ((tmp & SIC_LRB) == SIC_LRB) debug("LRB "); \
		if ((tmp & SIC_AAS) == SIC_AAS) debug("AAS "); \
		if ((tmp & SIC_LAB) == SIC_LAB) debug("LAB "); \
		if ((tmp & SIC_BBN) == SIC_BBN) debug("BBN "); \
		debug(")\n"); \
	} while (0)

#define I2C_ABORT() do { \
	i2c_set_bus_num(i2c_bus); \
} while (0)


int i2c_probe(uchar chip)
{
	uchar c[2];

	return i2c_read(chip, 0, 0, &c[0], 2);
}

void i2c_init (int speed,int slaveaddr)
{
	// 	i2c_set_bus_speed (speed);
	i2c_set_bus_num (CFG_CONTROLLER_BUS_NO);
}

int vsc452_i2c_init (int bus_no)
{
	return i2c_set_bus_num (bus_no);
}

int i2c_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	uchar tmp;
	int i;
	
	if (len < 1 || alen < 0) {
		return -1;
	}

	/* if shared bus */
	if (i2c_bus == 2) {

		/* request shared bus */
		if (i2c_request_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}	
		
		/* wait for bus grant and if bus granted */
		if (i2c_grant_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}
		
		/* start short timer */
		fpga_start_short_timer();
	}
	
	/* Wait for bus to stop */
	WAIT_UNTIL((iread(SIC) & SIC_BBN) == SIC_BBN, I2C_TIMEOUT) {
		debug("I2C bus timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}		
		return -1;
	}

	/* Send start */
	iwrite(SID, SID_DATA((chip << 1) | ((alen == 0) ? 1 : 0)));
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STA | SIC_ACK);

	/* Wait for NAK */
	WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
		I2C_STAT(tmp);
		debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}		
		
		return -1;
	}

	if (tmp & SIC_LRB) {
		debug("I2C slave NACK, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}		
		return -1;
	}

	/* Send address */
	for (i = 0; i < alen; i++) {
		I2C_STAT(tmp);
		if ((tmp & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}		
			return -1;
		}

		iwrite(SID, SID_DATA(addr >> ((alen - i - 1) * 8)));
		WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
			I2C_STAT(tmp);
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}			
			return -1;
		}
	}

	if (alen > 0) {
		/* Send chip id */
		iwrite(SID, SID_DATA((chip << 1) | 1));
		/* Send repeated start */
		iwrite(SIC, SIC_ES_DATA | SIC_STA | SIC_ACK);
	}


	for (i = 0; i < len; i++) {
		buffer[i] = 0x80 + i;
	}
	for (i = -1; i < (len - 1); i++) {
		WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
			debug("I2C read timeout, bus = %d, chip = 0x%02x, offset %d\n", i2c_bus, chip, i);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}				
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			/* Error! */
			iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
			debug("I2C read error, bus = %d, chip = 0x%02x, offset %d\n", i2c_bus, chip, i);
			tmp = iread(SID);	/* Throw-away data */
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}				
			return -1;
		}

		tmp = SID_DATA(iread(SID));
		if (i >= 0) {
			buffer[i] = tmp;
		}
	}

	iwrite(SIC, SIC_ES_DATA);
	tmp = SID_DATA(iread(SID));
	if (i >= 0) {
		buffer[i] = tmp;
	}

	WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
		debug("I2C finish error, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}			
		return -1;
	}

	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
	tmp = SID_DATA(iread(SID));
	buffer[i] = tmp;

	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_ACK);

	/* if shared bus */
	if (i2c_bus == 2) {
		/* release shared bus */
		i2c_release_shared_bus(shared_bus_seg);
	}
	
	return 0;
}

int i2c_write(uchar chip, uint addr, int alen, uchar * buffer, int len)
{
	int i;
	uchar tmp;

	/* if shared bus */
	if (i2c_bus == 2) {

		/* request shared bus */
		if (i2c_request_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}	
		
		/* wait for bus grant and if bus granted */
		if (i2c_grant_shared_bus(shared_bus_seg) != STATUS_SUCCESS) {
			return -1;
		}
		
		/* (re)start short timer */
		fpga_start_short_timer();
	}
	
	/* Wait for bus to stop */
	WAIT_UNTIL((iread(SIC) & SIC_BBN) == SIC_BBN, I2C_TIMEOUT) {
		debug("I2C bus timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}
		return -1;
	}

	/* Send start */
	iwrite(SID, SID_DATA((chip << 1) | 0));
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STA | SIC_ACK);

	/* Wait for NAK */
	WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
		I2C_STAT(tmp);
		debug("I2C NAK timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		I2C_ABORT();
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}
		return -1;
	}

	if (tmp & SIC_LRB) {
		debug("I2C slave NACK, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
		iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);
		/* if shared bus */
		if (i2c_bus == 2) {
			/* release shared bus */
			i2c_release_shared_bus(shared_bus_seg);
		}
		return -1;
	}

	/* Send address */
	for (i = 0; i < alen; i++) {
		iwrite(SID, SID_DATA(addr >> ((alen - i - 1) * 8)));

		/* (re)start short timer */
		if (i2c_bus == 2) { fpga_start_short_timer(); }
		
		WAIT_UNTIL(((tmp = iread(SIC)) & SIC_PIN) == 0, I2C_TIMEOUT) {
			I2C_STAT(tmp);
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}
			return -1;
		}
	}

	/* Send data */
	for (i = 0; i < len; i++) {
		iwrite(SID, SID_DATA(buffer[i]));

		/* (re)start short timer */
		if (i2c_bus == 2) { fpga_start_short_timer(); }
		
		WAIT_UNTIL((iread(SIC) & SIC_PIN) == 0, I2C_TIMEOUT) {
			debug("I2C address timeout, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}
			return -1;
		}

		if ((iread(SIC) & SIC_LRB) != 0) {
			debug("I2C slave missing, bus = %d, chip = 0x%02x\n", i2c_bus, chip);
			I2C_ABORT();
			/* if shared bus */
			if (i2c_bus == 2) {
				/* release shared bus */
				i2c_release_shared_bus(shared_bus_seg);
			}
			return -1;
		}
	}

	/* Send stop */
	iwrite(SIC, SIC_PIN | SIC_ES_DATA | SIC_STO | SIC_ACK);

	/* if shared bus */
	if (i2c_bus == 2) {
		/* release shared bus */
		i2c_release_shared_bus(shared_bus_seg);
	}
	
	return 0;
}

/*-------------------------------------------------------------------------
	Copyright (C) 2008-20013 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/

int supfpga_mm_reg8_read(uint16_t reg, uchar *read_buf);
int supfpga_mm_reg8_write(uint16_t reg, uchar write_buf);

int i2c_request_shared_bus(uchar segnum);
int i2c_grant_shared_bus(uchar segnum);
int i2c_release_shared_bus(uchar segnum);

int fpga_start_short_timer(void);
int fpga_set_local_short_timeout(uchar timeout);
int fpga_set_local_long_timeout(uchar timeout);

//#define DEBUG
//#define DEBUG_VERBOSE

//#define DEBUG_SFPGA
#define AVOID_ERROR_LOG_SPAM

#ifdef DEBUG_SFPGA
void dump_support_fpga_regs(void);
#endif

int
supfpga_mm_reg8_read(uint16_t reg, uchar *read_buf) {
	
	*read_buf = (uchar)READ_SUPFPGA_MMAP_REG(reg);
	return STATUS_SUCCESS;
}

int
supfpga_mm_reg16_read(uint16_t reg, uint16_t *read_buf) {
	
	*read_buf = (uint16_t)READ_SUPFPGA_MMAP_REG(reg);
	return STATUS_SUCCESS;
}

int
supfpga_mm_reg8_write(uint16_t reg, uchar write_buf) {

	WRITE_SUPFPGA_MMAP_REG(reg, (uint16_t)write_buf);
	return STATUS_SUCCESS;
}

int 
i2c_request_shared_bus(uchar segnum) {

	uchar val_mlbr, val_mrbr, val_mler, val_mrer;
	int retval=STATUS_SUCCESS;

#ifdef DEBUG_VERBOSE
	debug("i2c_request_shared_bus() ENTER: segnum=%d\n", (int)segnum);
#endif

	/* set local long timeout value */
	fpga_set_local_long_timeout(4); 
	/* set short timer */
	fpga_set_local_short_timeout(255);
	/* start short timer */
	fpga_start_short_timer();

	/* read local bus register (MLBR) */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* read remote bus register (MRBR) for debugging */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);

	/* if local bus request already asserted (MLBR[4] set) */
	if (val_mlbr & MLBR_LOC_BUS_REQ) {
		/* log warning */
		debug("  i2c_request_shared_bus(): Local bus request already asserted, segnum=%d\n", (int)segnum);

		/* clear local bus request */
		supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (val_mlbr & ~MLBR_LOC_BUS_REQ));

		/* again read local bus request */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

		/* if local bus request is still asserted (MLBR[4] set) */
		if (val_mlbr & MLBR_LOC_BUS_REQ) {
			/* log error and return */
			debug("i2c_request_shared_bus() FAIL: Local bus request still asserted after deasserting, segnum=%d, MLBR=0x%02x\n", 
				(int)segnum, val_mlbr);
			return STATUS_INACCESSIBLE;
		} 
	}

	/* read local error register (MLER) */
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	/* read remote error register (MRER) */
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
	
	/* if any local or remote errors */
	if ((val_mler & CLEARABLE_ERROR_MASK) || (val_mrer & CLEARABLE_ERROR_MASK)) {
		/* log errors */
		debug("  i2c_request_shared_bus(): Local and/or remote errors detected... attempting to clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);

		/* if any local errors */
		if (val_mler & CLEARABLE_ERROR_MASK) {
			/* clear errors from local error register (note: clear is done be setting bit) */
			supfpga_mm_reg8_write(SUPFPGA_MLER_REG, val_mler);
			/* reread local error register (MLER) */
			supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
		}

		/* if any remote errors */
		if (val_mrer & CLEARABLE_ERROR_MASK) {
			/* clear errors from remote error register (note: clear is done be setting bit) */
			supfpga_mm_reg8_write(SUPFPGA_MRER_REG, val_mrer);
			/* reread remote error register (MRER) */
			supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
		}
			
		/* if we still have local or remote errors after clearing */
		if ((val_mler != NO_ERROR) || (val_mrer != NO_ERROR)) {
			debug("i2c_request_shared_bus() FAIL: Local and/or remote errors detected after clearing, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
				(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);

			return STATUS_INACCESSIBLE;
		}
	}

	/* if local bus already active (MLBR[5]set) */
	if (val_mlbr & MLBR_LOC_BUS_ACT) {
		debug("i2c_request_shared_bus() FAIL: Local bus already active, segnum=%d, MLBR=0x%02x, MRBR=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr);
		return STATUS_INACCESSIBLE;
	} 

	/* set request bus number (segment number) */
	/* set the access type to read/write */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (segnum | MLBR_LOC_ACC_TYPE));
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* request bus */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (MLBR_LOC_BUS_REQ | val_mlbr));

	if (retval == STATUS_SUCCESS) {
#ifdef DEBUG_VERBOSE	
		/* read local bus register (MLBR) */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
		debug("i2c_request_shared_bus() SUCCESS: segnum=%d, MLBR=0x%02x, MRBR=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr);
#endif
	}
	return retval;
}

int 
i2c_grant_shared_bus(u8 segnum) {

	u8 val_mlbr, val_mrbr, val_mler, val_mrer, long_timeout;
	int retval=STATUS_SUCCESS;
	int time_left;
	static u8 last_mler=0, last_mrer=0;
	int timeout=0;
	int log_error=1; 

	/* read local bus register (MLBR) */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* read remote bus register (MRBR) */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);

	/* read long timeout value */
	supfpga_mm_reg8_read(SUPFPGA_MLTR_REG, &long_timeout);

	/* timeout in ms */	
	time_left = (long_timeout * 1000);
	
#ifdef DEBUG_VERBOSE	
	debug(KERN_INFO "i2c_grant_shared_bus() ENTER: segnum=%d\n", (int)segnum);
#endif

	/* while bus is not active (MLBR[5] clear) */
	while ((0 == (val_mlbr & MLBR_LOC_BUS_ACT))) {

		/* if timeout */
		if (time_left <= 0) {
			timeout = 1;
			break;
		}

		/* pet short timer */
		fpga_start_short_timer();
		
		/* Read local errors(MLER) and remote errors(MRER) */
		supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
		supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
		
		/* if any error detected (local, remote, clearable, unclearable) */
		if ((val_mler != NO_ERROR) || (val_mrer != NO_ERROR)) {

#ifdef AVOID_ERROR_LOG_SPAM
			/* if un-clearable errors only (invalid serial exchange bus AND local buffers not ready)*/
			if (((val_mler & CLEARABLE_ERROR_MASK) == NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) == NO_ERROR)) {

				 /* Some errors are not clearable and can potentially cause log spam.  
				 Avoid log spam by only logging once everytime error asserts */
				if ( /*((val_mler & 0x80) && (last_mler & 0x80)) || RESERVED */ 
					 ((val_mrer & 0x80) && (last_mrer & 0x80)) ||
					 ((val_mler & 0x40) && (last_mler & 0x40)) ||
					 ((val_mrer & 0x40) && (last_mrer & 0x40)) ||
					 ((val_mler & 0x20) && (last_mler & 0x20)) ||
					 ((val_mrer & 0x20) && (last_mrer & 0x20)) ) {
						log_error = 0; /* this error has already been logged, so skip logging */
				}
			}
#endif
			/* log the error (to avoid log spam from un-clearable errors 
			we only log if error has changed since last loop) */
			if (log_error) {
				debug("  i2c_grant_shared_bus(): Local and/or remote errors... attempting to clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
			}

#ifdef DEBUG_SFPGA
			dump_support_fpga_regs();
			retval = STATUS_INACCESSIBLE;
			break; /* note: request will be released before we return from function */
#endif
			
			/* release bus before attempting to clear errors */
			i2c_release_shared_bus(segnum);

			/* if mler errors attempt to clear */
			if ((val_mler & CLEARABLE_ERROR_MASK) != NO_ERROR) {
				supfpga_mm_reg8_write(SUPFPGA_MLER_REG, val_mler);
				supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
			}
			/* if mrer errors attempt to clear */
			if ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR) {
				supfpga_mm_reg8_write(SUPFPGA_MRER_REG, val_mrer);
				supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
			}

			/* re-request bus request */
			if (i2c_request_shared_bus(segnum) != 0) {
				debug("  i2c_grant_shared_bus(): Re-request of bus after error detection failed, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
				retval = STATUS_INACCESSIBLE;
				break;
			}

			/* if still have errors... log error */
			if (((val_mler & CLEARABLE_ERROR_MASK) != NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR)) {
			
				debug(KERN_WARNING "  i2c_grant_shared_bus(): Local and/or remote errors after clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);

				retval = STATUS_INACCESSIBLE;
				break;
			}
		}

		last_mler = val_mler;
		last_mrer = val_mrer;

		/* wait a little bit before trying again */
		udelay((time_left > 1000) ? 1000 : time_left);
		time_left -= (time_left > 1000)  ? 1000 : time_left;

		/* re-read MLBR and MRBR registers (update val_mlbr and val_mrbr) */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
		supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	}
 
	/* if wait timer >= long timeout value */
	if (timeout) {
#ifdef DEBUG
		debug("  i2c_grant_shared_bus(): Timeout waiting for bus grant, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
#endif	
		retval = STATUS_INACCESSIBLE;
	}

	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);

	/* release bus request on failures */
	if (retval != STATUS_SUCCESS) {
#ifdef DEBUG_VERBOSE	
		debug(KERN_WARNING "i2c_grant_shared_bus() FAIL, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
#endif			
		i2c_release_shared_bus(segnum);
	}

	if (retval == STATUS_SUCCESS) {
		fpga_start_short_timer();
#ifdef DEBUG_VERBOSE	
		debug(KERN_INFO "i2c_grant_shared_bus() SUCCESS, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
#endif	
	} 
	
	return retval;
}

int
i2c_release_shared_bus(u8 segnum) {

	u8 val_mlbr;
#ifdef DEBUG_VERBOSE	
	u8 val_mrbr, val_mler, val_mrer;
#endif

	/* read local bus register */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* release bus request (clear MLBR[4]) */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (val_mlbr & ~MLBR_LOC_BUS_REQ));	

#ifdef DEBUG_VERBOSE
	/* log any errors (we will clear them on next access) */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
	if (((val_mler & CLEARABLE_ERROR_MASK)!= NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR)) {
		debug("i2c_release_shared_bus(): segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x\n", (int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer);
	}
#endif

	return STATUS_SUCCESS;
}


int 
fpga_start_short_timer() {

	static int cnt=0;
	(cnt == 2) ? cnt = 0 : cnt++;

	supfpga_mm_reg8_write(SUPFPGA_MTRR_REG, MTRR_START_STIMER);

	return STATUS_SUCCESS;
}

int 
fpga_set_local_short_timeout(u8 timeout) {
	
#ifdef DEBUG_VERBOSE	
	debug("fpga set local short timeout to %d millisecond(s)\n", (int)timeout);
#endif
	supfpga_mm_reg8_write(SUPFPGA_MSTR_REG, timeout);

	return STATUS_SUCCESS;
}

int
fpga_set_local_long_timeout(u8 timeout) {
	
	/* max allowed long timeout value is 15 */
	if (timeout > 15) {
#ifdef DEBUG_VERBOSE	
		debug("fpga set local long timeout to %d second(s) exceeds max of 15 seconds\n   forcing timeout to 15 seconds\n", (int)timeout);
#endif
		timeout = 15;
	}
	
#ifdef DEBUG_VERBOSE	
	debug("fpga set local long timeout to %d second(s)\n", (int)timeout);
#endif
	supfpga_mm_reg8_write(SUPFPGA_MLTR_REG, timeout);

	return STATUS_SUCCESS;
}

#ifdef DEBUG_SFPGA
void
dump_support_fpga_regs() {

	uint16_t val, usw;

	/* log support fpga registers */
	debug("  SUPPORT FPGA REG DUMP: \n");

	supfpga_mm_reg16_read(SUPFPGA_MCSR_REG, &val);
	debug("    MCSR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLBR_REG, &val);
	debug("    MLBR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MTRR_REG, &val);
	debug("    MTRR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MRBR_REG, &val);
	debug("    MRBR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLER_REG, &val);
	debug("    MLER=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MRER_REG, &val);
	debug("    MRER=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MSTR_REG, &val);
	debug("    MSTR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLTR_REG, &val);
	debug("    MLTR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG0_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG0_REG-2, &usw);
	debug("    C000=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG1_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG1_REG-2, &usw);
	debug("    C004=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG2_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG2_REG-2, &usw);
	debug("    C008=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG3_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG3_REG-2, &usw);
	debug("    C00C=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG4_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG4_REG-2, &usw);
	debug("    C010=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG5_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG5_REG-2, &usw);
	debug("    C014=0x%04x%04x\n", usw, val);
	
	return;
}			
#endif

